import os
import re
import glob
import copy
import warnings
from os.path import join, isfile
from subprocess import check_output
from pathlib import Path

import git
import pandas as pd
import numpy as np

from amhelpers.amhelpers import save_yaml
from amhelpers.amhelpers import seed_hash
from amhelpers.amhelpers import create_results_dir_from_config


def _extract_jobid(output):
    output = output.decode('utf-8')
    jobid = int(output.split('job')[1])
    return jobid


def update_nested_dict(nested_dict, key_string, value):
    """Update a nested dictionary with a value.
    
    Parameters
    ----------
    nested_dict : dict
        The nested dictionary.
    key_string : str
        A string of keys separated by '::'.
    value : any
        The value to be inserted into the dictionary.

    Examples
    --------
    >>> d = {'a': {'b': {'c': 1}}}
    >>> update_nested_dict(d, 'a::b::c', 2)
    >>> d
    {'a': {'b': {'c': 2}}}
    """
    keys = key_string.split('::')
    current = nested_dict
    for key in keys[:-1]:
        current = current.setdefault(key, {})
    current[keys[-1]] = value


def create_jobscript_from_template(
    template, experiment, experiment_path, estimator, jobname, jobdir=None,
    options={}
):
    with open(template, 'r') as f:
        text = f.read()
    
    text = text.replace('<EXPERIMENT>', experiment)
    text = text.replace('<EXPERIMENT_PATH>', experiment_path)
    text = text.replace('<ESTIMATOR>', estimator)
    
    for key, value in options.items():
        text = text.replace('<%s>' % key.upper(), value)
    
    if jobdir is None:
        jobdir = experiment_path
    
    jobscript_path = join(jobdir, jobname)
    with open(jobscript_path, 'w') as f:
        f.write(text)
    
    return jobscript_path


class Sweep:
    def __init__(
        self, config, estimators, job_template_training, 
        job_template_preprocessing, job_template_postprocessing,
        sweep_param=None, sweep_param_values=None, sweep_param_str=None,
        n_trials=5, n_hparams=10, include_default_hparams=True,
        options={'gpu': 'A40'}
    ):
        self.path, self.config = create_results_dir_from_config(
            config, suffix='sweep', update_config=True
        )
        
        self.estimators = estimators
        
        self.job_template_training = job_template_training
        self.job_template_preprocessing = job_template_preprocessing
        self.job_template_postprocessing = job_template_postprocessing
        
        self.sweep_param = sweep_param
        self.sweep_param_values = sweep_param_values
        self.sweep_param_str = sweep_param_str if sweep_param_str is not None \
            else sweep_param

        self.n_trials = n_trials
        self.n_hparams = n_hparams
        self.include_default_hparams = include_default_hparams
        
        self.options = options
    
    def _create_job_configs(self, default_config, sweep_dir_name='sweep'):
        trials_range = range(1, self.n_trials+1)
        hparams_range = range(self.n_hparams) if self.include_default_hparams \
            else range(1, self.n_hparams+1)

        for estimator in self.estimators:
            config_dir = join(self.path, 'configs', estimator)
            Path(config_dir).mkdir(parents=True, exist_ok=True)

            num_configs = len(os.listdir(config_dir))
            i_config = num_configs + 1

            for trial_seed in trials_range:
                for hparams_seed in hparams_range:
                    config = copy.deepcopy(default_config)
                    
                    config['data']['seed'] = trial_seed
                    config['hparams']['seed'] = hparams_seed
                    config['estimators']['seed'] = seed_hash(
                        config['experiment'], estimator, trial_seed,
                        hparams_seed
                    )
                    
                    results_path = join(
                        self.path, sweep_dir_name, f'trial_{trial_seed:02d}',
                        f'{estimator}_{hparams_seed:02d}'
                    )
                    config['results']['path'] = results_path
                    
                    Path(results_path).mkdir(parents=True)
                    save_yaml(config, results_path, 'config')

                    save_yaml(config, config_dir, f'config{i_config:03d}')
                    i_config += 1

    def prepare(self):
        # =====================================================================
        # Create a directory for log files and save Git info.
        # =====================================================================

        logs_dir = join(self.path, 'logs')
        Path(logs_dir).mkdir()

        try:
            repo = git.Repo(search_parent_directories=True)
            with open(join(self.path, 'repo.tar'), 'wb') as f:
                repo.archive(f)
            git_info = {
                'branch': repo.active_branch, 'hash': repo.head.object.hexsha
            }
            with open(join(self.path, 'gitinfo.txt'), 'w') as f:
                print(git_info, file=f)
        except Exception as e:
            warnings.warn(f"Getting Git info failed: {e}.")

        # =====================================================================
        # Create a config file for each job.
        # =====================================================================
        
        save_yaml(self.config, self.path, 'default_config')

        if self.sweep_param is not None:
            assert self.sweep_param_values is not None
            for value in self.sweep_param_values:
                config = copy.deepcopy(self.config)
                update_nested_dict(config, self.sweep_param, value)
                dir_name = f'sweep_{self.sweep_param_str}_{value}'
                self._create_job_configs(config, dir_name)
            self._n_jobs = len(self.sweep_param_values) * self.n_trials * self.n_hparams
        else:
            self._create_job_configs(self.config)
            self._n_jobs = self.n_trials * self.n_hparams
    
        # =====================================================================
        # Create jobscripts.
        # =====================================================================

        jobscripts_dir = join(self.path, 'jobscripts')
        Path(jobscripts_dir).mkdir()

        self.jobsripcts = {'pre': None, 'main': {}, 'post': None}

        experiment = self.config['experiment']
        kwargs = {
            'experiment': experiment, 'experiment_path': self.path,
            'jobdir': jobscripts_dir, 'options': self.options
        }
        
        jobscript_path = create_jobscript_from_template(
            template=self.job_template_preprocessing, estimator='',
            jobname='job_pre', **kwargs
        )
        self.jobsripcts['pre'] = jobscript_path

        for estimator in self.estimators:
            jobname = f'job_{estimator}'
            jobscript_path = create_jobscript_from_template(
                template=self.job_template_training, estimator=estimator,
                jobname=jobname, **kwargs
            )
            self.jobsripcts['main'][estimator] = jobscript_path

        jobscript_path = create_jobscript_from_template(
            template=self.job_template_postprocessing, estimator='',
            jobname='job_post', **kwargs
        )
        self.jobsripcts['post'] = jobscript_path
    
    def launch(self):
        main_dependencies = [self._submit_job(self.jobsripcts['pre'])]
        
        post_dependencies = []
        for estimator in self.estimators:
            jobid = self._submit_job(
                self.jobsripcts['main'][estimator], main_dependencies,
                self._n_jobs
            )
            post_dependencies.append(jobid)
        
        self._submit_job(self.jobsripcts['post'], post_dependencies)

    def _submit_job(self, jobscript_path, dependencies=None, n_jobs=1):
        command = ['sbatch']
        if dependencies is not None:
            dependencies = [str(d) for d in dependencies]
            dependencies = ':'.join(dependencies)
            command.append(f'--dependency=afterok:{dependencies}')
        if n_jobs > 1:
            command.append(f'--array=1-{n_jobs}')
        command.append(jobscript_path)
        return _extract_jobid(check_output(command))


class Postprocessing:
    def __init__(self, exp_path):
        self.exp_path = exp_path
        config_dir = join(exp_path, 'configs')
        self.estimators = os.listdir(config_dir)
    
    def _get_sweep_dirs(self):
        sweep_dirs = [
            join(self.exp_path, x) for x in os.listdir(self.exp_path)
            if x.startswith('sweep')
        ]
        return sorted(sweep_dirs)
    
    def _get_trial_dirs(self, sweep_dir):
        trial_dirs = [
            join(sweep_dir, x) for x in os.listdir(sweep_dir)
            if 'trial' in x
        ]
        return sorted(trial_dirs)

    def _sort_scores(self, path, sorter):
        all_sorted_scores = {}
        
        for estimator in self.estimators:
            experiments = [
                x for x in os.listdir(path) if x.startswith(estimator)
            ]

            csvs, scores = [], []
            
            for e in experiments:
                p = join(path, e, 'scores.csv')
                if not isfile(p):
                    print(f"File {p} not found.")
                    continue
                
                csv = pd.read_csv(p)
                if not 'subset' in csv.columns:
                    raise ValueError(f"Column 'subset' not found in {p}.")
                csv.insert(0, 'exp', e)
                csvs.append(csv)

                score = sorter(csv)
                if not isinstance(score, float):
                    raise TypeError(
                        "The output of the score sorting function must "
                        f"be a float, but got type {type(score).__name__}."
                    )
                scores.append(score)
            
            if len(csvs) > 0:
                scores = np.array(scores)
                if np.isnan(scores).any():
                    scores[np.isnan(scores)] = -np.inf
                csvs = [csvs[i] for i in np.argsort(scores)[::-1]]
                sorted_scores = pd.concat(csvs, ignore_index=True)
                all_sorted_scores[estimator] = sorted_scores
        
        return all_sorted_scores

    def _collect_best_scores(self, sorted_scores):
        best_scores_list = []

        for i_trial in range(len(sorted_scores)):
            for _estimator, scores in sorted_scores[i_trial].items():
                best_exp = scores.exp.iloc[0]
                best_score = scores.groupby('exp').get_group(best_exp)
                trial = pd.DataFrame({'trial': len(best_score) * [i_trial+1]})
                best_score = pd.concat((trial, best_score), axis=1)
                best_scores_list.append(best_score)
 
        return best_scores_list

    def _concat_scores(self, sorted_scores):
        best_scores = self._collect_best_scores(sorted_scores)
        scores = pd.concat(best_scores, ignore_index=True)
        return scores

    def collect_results(self, score_sorter=None):
        if score_sorter is None:
            score_sorter = lambda csv: csv[csv.subset=='valid']['auc'].item()

        scores = []
        for sweep_dir in self._get_sweep_dirs():
            sorted_scores = []
            for trial_dir in self._get_trial_dirs(sweep_dir):
                sorted_scores += [self._sort_scores(trial_dir, score_sorter)]
            
            sweep_scores = self._concat_scores(sorted_scores)
            
            sweep_dir_name = sweep_dir.split('/')[-1]
            if sweep_dir_name != 'sweep':
                sweep_param_value = sweep_dir_name.split('_')[-1]
                pattern = r'sweep_(.*?)_' + re.escape(sweep_param_value)
                match = re.search(pattern, sweep_dir_name)
                sweep_param = match.group(1)
                sweep_scores.insert(0, sweep_param, sweep_param_value)
            
            scores.append(sweep_scores)
        
        scores = pd.concat(scores, ignore_index=True)
        scores.to_csv(join(self.exp_path, 'scores.csv'), index=False)
    
    def remove_files(self):
        scores_path = join(self.exp_path, 'scores.csv')
        if not isfile(scores_path):
            raise FileNotFoundError(
                f"File {scores_path} is required but was not found."
            )
        scores = pd.read_csv(scores_path)

        dirs_to_keep = []
        if scores.columns[0] != 'trial':
            assert scores.columns[1:3].tolist() == ['trial', 'exp']
            sweep_param = scores.columns[0]
            for sweep_param_value, trial, exp in scores.iloc[:, :3].values:
                sweep = f'sweep_{sweep_param}_{sweep_param_value}'
                dirs_to_keep += [
                    join(self.exp_path, sweep, f'trial_{trial:02d}', exp)
                ]
        else:
            assert scores.columns[:2].tolist() == ['trial', 'exp']
            dirs_to_keep = [
                join(self.exp_path, 'sweep', f'trial_{trial:02d}', exp)
                for trial, exp in scores.iloc[:, :2].values
            ]

        all_dirs = []
        for sweep_dir in self._get_sweep_dirs():
            for trial_dir in self._get_trial_dirs(sweep_dir):
                for x in os.listdir(trial_dir):
                    all_dirs.append(join(trial_dir, x))

        dirs_to_delete = set(all_dirs) - set(dirs_to_keep)

        for d in dirs_to_delete:
            for f in glob.glob(join(d, '*.pt')):
                os.remove(f)
            for f in glob.glob(join(d, '*.pkl')):
                os.remove(f)
            for f in glob.glob(join(d, '*.json')):
                os.remove(f)



import numpy as np


def ece(y, y_probas, n_bins=10):
    """Expected calibration error.

    Parameters
    ----------
    y : NumPy array of shape (n_samples, n_classes) or (n_samples,)
        True labels. Either as label vector (1-D) or as one-hot encoded ground 
        truth array (2-D).
    y_probas : NumPy array of shape (n_samples, n_classes) or (n_samples,)
        Predicted probabilities.

    Returns
    -------
    ece : float
        ECE score.
    """
    if y_probas.ndim == 2:
        max_indices = y_probas.argmax(axis=-1)
        if y.ndim == 2:
            indices = np.expand_dims(max_indices, axis=-1)
            y = np.take_along_axis(y, indices, axis=-1).squeeze(axis=-1)
        elif not np.array_equal(y, y.astype(bool)):
            n_classes = y_probas.shape[1]
            assert set(y).issubset(range(n_classes))
            y = np.array([yi == i for yi, i in zip(y, max_indices)], dtype=bool)
        indices = np.expand_dims(max_indices, axis=-1)
        y_probas = np.take_along_axis(y_probas, indices, axis=-1).squeeze(axis=-1)
    else:
        assert y.ndim == 1 and np.array_equal(y, y.astype(bool))
        
    bins = np.linspace(0, 1, n_bins+1)
    bin_indices = np.digitize(y_probas, bins, right=False)  # Assuming no probabilities equal 1.0

    _ece = 0
    for i in range(1, n_bins+1):
        bin_mask = bin_indices == i
        if bin_mask.sum() == 0:
            continue
        y_probas_bin = y_probas[bin_mask]
        conf = np.mean(y_probas_bin)
        y_bin = y[bin_mask]     
        acc = np.mean(y_bin)
        _ece += len(y_bin) * np.abs(acc-conf)

    return _ece / len(y_probas)


def sce(y_one_hot, y_probas, n_bins=10):
    """Static calibration error."""
    _sce = 0
    for yp, y in zip(y_probas.T, y_one_hot.T):
        _sce += ece(y, yp, n_bins)
    return _sce / y_probas.shape[1]



import os
import copy
from pathlib import Path

from .amhelpers import (
    get_class_from_str,
    create_object_from_dict,
    load_yaml
)


NET_PARAMS = [
    'module',
    'criterion',
    'optimizer',
    'lr',
    'max_epochs',
    'batch_size',
    'iterator_train',
    'iterator_valid',
    'dataset',
    'train_split',
    'callbacks',
    'predict_nonlinearity',
    'warm_start',
    'verbose',
    'device'
]


def _check_value(value):
    if isinstance(value, (int, float)):
        return value
    elif isinstance(value, bool):
        return value
    elif isinstance(value, str):
        return value
    elif isinstance(value, list):
        return [_check_value(v) for v in value]
    elif isinstance(value, dict):
        if 'type' in value:
            is_called = value.pop('is_called')
            if is_called:
                has_nested_dict = any(
                    [isinstance(v, dict) and 'type' in v for v in value.values()]
                )
                if has_nested_dict:
                    return create_object_from_dict(
                        {k: _check_value(v) for k, v in value.items()}
                    )
                else:
                    return create_object_from_dict(value)
            else:
                type_str = value['type']
                return get_class_from_str(type_str)
        else:
            return {k: _check_value(v) for k, v in value.items()}
    elif value is None:
        return value
    else:
        raise ValueError("The value of type {} is unknown.".format(type(value)))


def _get_object_and_parameters(name, default_params, specified_params):
    default_params.pop('is_called', None)
    specified_params.pop('is_called', None)

    prefix = name + '__'

    if 'type' in specified_params:
        # Take everything from specified_params.
        out = {name: get_class_from_str(specified_params.pop('type'))}
        out.update(
            {prefix+k: _check_value(v) for k, v in specified_params.items()}
        )
    elif 'type' in default_params:
        # Replace default values if they exist in specified_params.
        out = {name: get_class_from_str(default_params.pop('type'))}
        new = {}
        for k in default_params.keys():
            if k in specified_params:
                v = specified_params.pop(k)
                new[prefix+k] = _check_value(v)
            else:
                new[prefix+k] = _check_value(default_params[k])
        out.update(new)
        # Add values that exist in specified_params but not in default_params.
        out.update({prefix+k: _check_value(v) for k, v in specified_params.items()})
    else:
        out = {}
        for k in set(list(default_params.keys()) + list(specified_params.keys())):
            out[prefix+k] = _check_value(specified_params[k]) if k in specified_params else _check_value(default_params[k])

    return out


def get_net_params(default, specified):
    """Get parameters for a skorch neural net.

    Parameters
    ----------
    default : dict
        Default parameter values on the format "parameter name: value".
    specified : dict
        Specified parameter values on the format "parameter name: value". Will replace the default values.

    Returns
    -------
    params : dict
        All model parameters.
    """
    default = copy.deepcopy(default)
    specified = copy.deepcopy(specified)
    
    params = {}
    for param in NET_PARAMS:
        try:
            default_value = default[param]
        except KeyError:
            default_value = {}
        try:
            specified_value = specified[param]
        except KeyError:
            specified_value = {}

        if param in [
            'module',
            'criterion',
            'optimizer',
            'iterator_train',
            'iterator_valid',
            'dataset'
        ]:
            params.update(
                _get_object_and_parameters(param, default_value, specified_value)
            )
        else:
            if param in specified:
                params[param] = _check_value(specified[param])
            elif param in default:
                params[param] = _check_value(default_value)
            else:
                # Use skorch default.
                pass

    return params


def _change_to_local_paths(d, cluster_project_path, local_project_path):
    out = {}
    for k, v in d.items():
        if isinstance(v, dict):
            recursive = {
                k: _change_to_local_paths(
                    v,
                    cluster_project_path,
                    local_project_path
                )
            }
            out.update(recursive)
        elif isinstance(k, str) and 'path' in k and isinstance(v, str):
            out[k] = v.replace(
                cluster_project_path,
                local_project_path
            )
        else:
            out[k] = v
    return out


def load_config(config_path):
    config = load_yaml(config_path)
    try:
        local_home_path = os.environ['LOCAL_HOME_PATH']
        cluster_project_path = os.environ['CLUSTER_PROJECT_PATH']
        local_project_path = os.environ['LOCAL_PROJECT_PATH']
        home_path = str(Path.home())
        if home_path == local_home_path:
            return _change_to_local_paths(
                config,
                cluster_project_path,
                local_project_path
            )
    except KeyError:
        return config



import pydoc
import copy
import datetime
import pickle
import hashlib
from os.path import join
from pathlib import Path

import yaml


def create_object_from_dict(d, **default_kwargs):
    """Create an object from a dictionary.

    Inspired by `iglovikov_helper_functions.config_parsing.utils`. [1]_

    Parameters
    ----------
    d : dict
        Dictionary to create the object from. Must have a key "type" which specifies the class of the object.
        Any additional key-value pairs will be passed to the constructor when creating the object.
    **default_kwargs : additional arguments
        Default parameters to be passed to the constructor, if they do not exist in `d`.

    Returns
    -------
    object
        The initialized object.

    References
    ----------
    .. [1] https://github.com/ternaus/iglovikov_helper_functions/
    """
    if not isinstance(d, dict):
        raise TypeError("'d' should be of type 'dict'.")
    if not 'type' in d:
        raise KeyError("'d' should have a key 'type'.")
    kwargs = d.copy()
    class_type = kwargs.pop('type')
    for name, value in default_kwargs.items():
        kwargs.setdefault(name, value)
    return pydoc.locate(class_type)(**kwargs)


def get_class_from_str(s):
    """Get the uninitialized class `s`.

    Parameters
    ----------
    s : str
        The name of the class.

    Returns
    -------
    class
        The uninitialized class `s`.
    """
    return pydoc.locate(s)


def yield_nested_dict_values(d):
    """Yield (non-dict) values from a nested dict.

    Code from [2]_.

    Parameters
    ----------
    d : dict
        Nested dictionary.

    Returns
    -------
    generator
        Iterator yielding all (non-dict) values from `d` and its child dictionaries.

    References
    ----------
    .. [2] https://tutorial.eyehunts.com/python/python-get-all-values-from-nested-dictionary-example-code/
    """
    for v in d.values():
        if isinstance(v, dict):
            yield from yield_nested_dict_values(v)
        else:
            yield v


def save_pickle(data, path, filename):
    Path(path).mkdir(parents=True, exist_ok=True)
    with open(join(path, filename + '.pickle'), 'wb') as f:
        pickle.dump(data, f)


def save_yaml(data, path, filename, **kwargs):
    Path(path).mkdir(parents=True, exist_ok=True)
    with open(join(path, filename + '.yaml'), 'w') as f:
        yaml.dump(data, f, default_flow_style=False, **kwargs)


def load_yaml(path):
    with open(path) as file:
        yaml_file = yaml.safe_load(file)
    return yaml_file


def seed_hash(*args):
    """Derive an integer hash from `args` to use as a random seed."""
    args_str = str(args)
    return int(hashlib.md5(args_str.encode('utf-8')).hexdigest(), 16) % (2**31)


def create_results_dir_from_config(
    config,
    suffix=None,
    update_config=False
):
    """Create a directory for storing results.

    Parameters
    ----------
    config : dict
        Configuration dictionary. Must have a key 'results' with a key 'path' 
        specifying the path to the results directory.
    suffix : str, optional
        Suffix to append to the results directory name.
    update_config : bool, optional
        Whether to update the configuration dictionary with the new results path.
    """
    time_stamp = datetime.datetime.now().strftime('%Y%m%d_%H%M')
    if suffix is not None:
        time_stamp += '_' + suffix
    results_path = join(config['results']['path'], time_stamp)
    Path(results_path).mkdir(parents=True, exist_ok=True)
    
    if update_config:
        config = copy.deepcopy(config)
        config['results']['path'] = results_path
        return results_path, config
    else:
        return results_path
